@@ -222,6 +222,7 @@ cluster_read_sock_resp(RedisSock *redis_sock, REDIS_REPLY_TYPE type,
                 r->str = estrndup(line_reply, len);
                 r->len = len;
             }
+            REDIS_FALLTHROUGH;
         case TYPE_ERR:
             return r;
         case TYPE_BULK:
@@ -265,15 +266,14 @@ static int cluster_send_direct(RedisSock *redis_sock, char *cmd, int cmd_len,
     /* Connect to the socket if we aren't yet and send our command, validate the reply type, and consume the first line */
     if (!CLUSTER_SEND_PAYLOAD(redis_sock,cmd,cmd_len) ||
         !CLUSTER_VALIDATE_REPLY_TYPE(redis_sock, type) ||
-        !php_stream_gets(redis_sock->stream, buf, sizeof(buf))) return -1;
+        !redis_sock_gets_raw(redis_sock, buf, sizeof(buf))) return -1;
 
     /* Success! */
     return 0;
 }
 
 static int cluster_send_asking(RedisSock *redis_sock) {
-    return cluster_send_direct(redis_sock, RESP_ASKING_CMD,
-        sizeof(RESP_ASKING_CMD)-1, TYPE_LINE);
+    return cluster_send_direct(redis_sock, ZEND_STRL(RESP_ASKING_CMD), TYPE_LINE);
 }
 
 /* Send READONLY to a specific RedisSock unless it's already flagged as being
@@ -286,8 +286,7 @@ static int cluster_send_readonly(RedisSock *redis_sock) {
     if (redis_sock->readonly) return 0;
 
     /* Return success if we can send it */
-    ret = cluster_send_direct(redis_sock, RESP_READONLY_CMD,
-        sizeof(RESP_READONLY_CMD) - 1, TYPE_LINE);
+    ret = cluster_send_direct(redis_sock, ZEND_STRL(RESP_READONLY_CMD), TYPE_LINE);
 
     /* Flag this socket as READONLY if our command worked */
     redis_sock->readonly = !ret;
@@ -298,9 +297,8 @@ static int cluster_send_readonly(RedisSock *redis_sock) {
 
 /* Send MULTI to a specific ReidsSock */
 static int cluster_send_multi(redisCluster *c, short slot) {
-    if (cluster_send_direct(SLOT_SOCK(c,slot), RESP_MULTI_CMD,
-        sizeof(RESP_MULTI_CMD) - 1, TYPE_LINE) == 0)
-    {
+    if (cluster_send_direct(SLOT_SOCK(c,slot), ZEND_STRL(RESP_MULTI_CMD), TYPE_LINE) == 0) {
+        c->flags->txBytes += sizeof(RESP_MULTI_CMD) - 1;
         c->cmd_sock->mode = MULTI;
         return 0;
     }
@@ -315,8 +313,7 @@ PHP_REDIS_API int cluster_send_exec(redisCluster *c, short slot) {
     int retval;
 
     /* Send exec */
-    retval = cluster_send_slot(c, slot, RESP_EXEC_CMD, sizeof(RESP_EXEC_CMD)-1,
-        TYPE_MULTIBULK);
+    retval = cluster_send_slot(c, slot, ZEND_STRL(RESP_EXEC_CMD), TYPE_MULTIBULK);
 
     /* We'll either get a length corresponding to the number of commands sent to
      * this node, or -1 in the case of EXECABORT or WATCH failure. */
@@ -327,8 +324,7 @@ PHP_REDIS_API int cluster_send_exec(redisCluster *c, short slot) {
 }
 
 PHP_REDIS_API int cluster_send_discard(redisCluster *c, short slot) {
-    if (cluster_send_direct(SLOT_SOCK(c,slot), RESP_DISCARD_CMD,
-        sizeof(RESP_DISCARD_CMD)-1, TYPE_LINE))
+    if (cluster_send_direct(SLOT_SOCK(c,slot), ZEND_STRL(RESP_DISCARD_CMD), TYPE_LINE))
     {
         return 0;
     }
@@ -358,7 +354,7 @@ static void cluster_dist_free_ht(zval *p) {
 }
 
 /* Spin up a HashTable that will contain distribution lists */
-HashTable *cluster_dist_create() {
+HashTable *cluster_dist_create(void) {
     HashTable *ret;
 
     ALLOC_HASHTABLE(ret);
@@ -374,7 +370,7 @@ void cluster_dist_free(HashTable *ht) {
 }
 
 /* Create a clusterDistList object */
-static clusterDistList *cluster_dl_create() {
+static clusterDistList *cluster_dl_create(void) {
     clusterDistList *dl;
 
     dl        = emalloc(sizeof(clusterDistList));
@@ -454,7 +450,7 @@ void cluster_dist_add_val(redisCluster *c, clusterKeyVal *kv, zval *z_val
     // Serialize our value
     val_free = redis_pack(c->flags, z_val, &val, &val_len);
 
-    // Attach it to the provied keyval entry
+    // Attach it to the provided keyval entry
     kv->val = val;
     kv->val_len = val_len;
     kv->val_free = val_free;
@@ -472,7 +468,7 @@ void cluster_multi_add(clusterMultiCmd *mc, char *data, int data_len) {
     redis_cmd_append_sstr(&(mc->args), data, data_len);
 }
 
-/* Finalize a clusterMutliCmd by constructing the whole thing */
+/* Finalize a clusterMultiCmd by constructing the whole thing */
 void cluster_multi_fini(clusterMultiCmd *mc) {
     mc->cmd.len = 0;
     redis_cmd_init_sstr(&(mc->cmd), mc->argc, mc->kw, mc->kw_len);
@@ -529,6 +525,10 @@ unsigned short cluster_hash_key(const char *key, int len) {
     return crc16((char*)key+s+1,e-s-1) & REDIS_CLUSTER_MOD;
 }
 
+unsigned short cluster_hash_key_zstr(zend_string *key) {
+    return cluster_hash_key(ZSTR_VAL(key), ZSTR_LEN(key));
+}
+
 /* Grab the current time in milliseconds */
 long long mstime(void) {
     struct timeval tv;
@@ -683,7 +683,7 @@ static int cluster_map_slots(redisCluster *c, clusterReply *r) {
     clusterReply *r2, *r3;
     unsigned short port;
     char *host, key[1024];
-
+    zend_hash_clean(c->nodes);
     for (i = 0; i < r->elements; i++) {
         // Inner response
         r2 = r->element[i];
@@ -708,27 +708,27 @@ static int cluster_map_slots(redisCluster *c, clusterReply *r) {
         if ((pnode = zend_hash_str_find_ptr(c->nodes, key, klen)) == NULL) {
             master = cluster_node_create(c, host, hlen, port, low, 0);
             zend_hash_str_update_ptr(c->nodes, key, klen, master);
+
+            // Attach slaves first time we encounter a given master in order to avoid registering the slaves multiple times
+            for (j = 3; j< r2->elements; j++) {
+                r3 = r2->element[j];
+                if (!VALIDATE_SLOTS_INNER(r3)) {
+                    return -1;
+                }
+
+                // Skip slaves where the host is ""
+                if (r3->element[0]->len == 0) continue;
+
+                // Attach this node to our slave
+                slave = cluster_node_create(c, r3->element[0]->str,
+                    (int)r3->element[0]->len,
+                    (unsigned short)r3->element[1]->integer, low, 1);
+                cluster_node_add_slave(master, slave);
+            }
         } else {
             master = pnode;
         }
 
-        // Attach slaves
-        for (j = 3; j< r2->elements; j++) {
-            r3 = r2->element[j];
-            if (!VALIDATE_SLOTS_INNER(r3)) {
-                return -1;
-            }
-
-            // Skip slaves where the host is ""
-            if (r3->element[0]->len == 0) continue;
-
-            // Attach this node to our slave
-            slave = cluster_node_create(c, r3->element[0]->str,
-                (int)r3->element[0]->len,
-                (unsigned short)r3->element[1]->integer, low, 1);
-            cluster_node_add_slave(master, slave);
-        }
-
         // Attach this node to each slot in the range
         for (j = low; j<= high; j++) {
             c->master[j] = master;
@@ -833,7 +833,7 @@ PHP_REDIS_API redisCluster *cluster_create(double timeout, double read_timeout,
     c->err = NULL;
 
     /* Set up our waitms based on timeout */
-    c->waitms  = (long)(1000 * timeout);
+    c->waitms  = (long)(1000 * (timeout + read_timeout));
 
     /* Allocate our seeds hash table */
     ALLOC_HASHTABLE(c->seeds);
@@ -1097,7 +1097,7 @@ PHP_REDIS_API int cluster_map_keyspace(redisCluster *c) {
                 memset(c->master, 0, sizeof(redisClusterNode*)*REDIS_CLUSTER_SLOTS);
             }
         }
-        redis_sock_disconnect(seed, 0);
+        redis_sock_disconnect(seed, 0, 1);
         if (mapped) break;
     } ZEND_HASH_FOREACH_END();
 
@@ -1120,6 +1120,10 @@ static int cluster_set_redirection(redisCluster* c, char *msg, int moved)
 {
     char *host, *port;
 
+    /* The Redis Cluster specification suggests clients do not update
+     * their slot mapping for an ASK redirection, only for MOVED */
+    if (moved) c->redirections++;
+
     /* Move past "MOVED" or "ASK */
     msg += moved ? MOVED_LEN : ASK_LEN;
 
@@ -1147,7 +1151,7 @@ static int cluster_set_redirection(redisCluster* c, char *msg, int moved)
  * redirection, parsing out slot host and port so the caller can take
  * appropriate action.
  *
- * In the case of a non MOVED/ASK error, we wlll set our cluster error
+ * In the case of a non MOVED/ASK error, we will set our cluster error
  * condition so GetLastError can be queried by the client.
  *
  * This function will return -1 on a critical error (e.g. parse/communication
@@ -1160,8 +1164,8 @@ static int cluster_check_response(redisCluster *c, REDIS_REPLY_TYPE *reply_type)
     CLUSTER_CLEAR_ERROR(c);
     CLUSTER_CLEAR_REPLY(c);
 
-    if (-1 == redis_check_eof(c->cmd_sock, 1) ||
-       EOF == (*reply_type = php_stream_getc(c->cmd_sock->stream)))
+    if (-1 == redis_check_eof(c->cmd_sock, 1, 1) ||
+       EOF == (*reply_type = redis_sock_getc(c->cmd_sock)))
     {
         return -1;
     }
@@ -1169,31 +1173,22 @@ static int cluster_check_response(redisCluster *c, REDIS_REPLY_TYPE *reply_type)
     // In the event of an ERROR, check if it's a MOVED/ASK error
     if (*reply_type == TYPE_ERR) {
         char inbuf[4096];
+        size_t nbytes;
         int moved;
 
         // Attempt to read the error
-        if (!php_stream_gets(c->cmd_sock->stream, inbuf, sizeof(inbuf))) {
+        if (!redis_sock_get_line(c->cmd_sock, inbuf, sizeof(inbuf), &nbytes)) {
             return -1;
         }
 
         // Check for MOVED or ASK redirection
         if ((moved = IS_MOVED(inbuf)) || IS_ASK(inbuf)) {
-            /* The Redis Cluster specification suggests clients do not update
-             * their slot mapping for an ASK redirection, only for MOVED */
-            if (moved) c->redirections++;
-
             /* Make sure we can parse the redirection host and port */
-            if (cluster_set_redirection(c,inbuf,moved) < 0) {
-                return -1;
-            }
-
-            /* We've been redirected */
-            return 1;
-        } else {
-            // Capture the error string Redis returned
-            cluster_set_err(c, inbuf, strlen(inbuf)-2);
-            return 0;
+            return !cluster_set_redirection(c, inbuf, moved) ? 1 : -1;
         }
+        // Capture the error string Redis returned
+        cluster_set_err(c, inbuf, strlen(inbuf)-2);
+        return 0;
     }
 
     // Fetch the first line of our response from Redis.
@@ -1223,13 +1218,13 @@ PHP_REDIS_API void cluster_disconnect(redisCluster *c, int force) {
         if (node == NULL) continue;
 
         /* Disconnect from the master */
-        redis_sock_disconnect(node->sock, force);
+        redis_sock_disconnect(node->sock, force, 1);
 
         /* We also want to disconnect any slave connections so they will be pooled
          * in the event we are using persistent connections and connection pooling. */
         if (node->slaves) {
             ZEND_HASH_FOREACH_PTR(node->slaves, slave) {
-                redis_sock_disconnect(slave->sock, force);
+                redis_sock_disconnect(slave->sock, force, 1);
             } ZEND_HASH_FOREACH_END();
         }
     } ZEND_HASH_FOREACH_END();
@@ -1268,9 +1263,7 @@ static int cluster_dist_write(redisCluster *c, const char *cmd, size_t sz,
 
         /* If we're not on the master, attempt to send the READONLY command to
          * this slave, and skip it if that fails */
-        if (nodes[i] == 0 || redis_sock->readonly ||
-            cluster_send_readonly(redis_sock) == 0)
-        {
+        if (nodes[i] == 0 || cluster_send_readonly(redis_sock) == 0) {
             /* Attempt to send the command */
             if (CLUSTER_SEND_PAYLOAD(redis_sock, cmd, sz)) {
                 c->cmd_sock = redis_sock;
@@ -1396,7 +1389,7 @@ static void cluster_update_slot(redisCluster *c) {
     /* Do we already have the new slot mapped */
     if (c->master[c->redir_slot]) {
         /* No need to do anything if it's the same node */
-        if (!CLUSTER_REDIR_CMP(c)) {
+        if (!CLUSTER_REDIR_CMP(c, SLOT_SOCK(c,c->redir_slot))) {
             return;
         }
 
@@ -1407,6 +1400,22 @@ static void cluster_update_slot(redisCluster *c) {
             /* Just point to this slot */
             c->master[c->redir_slot] = node;
         } else {
+            /* If the redirected node is a replica of the previous slot owner, a failover has taken place.
+            We must then remap the cluster's keyspace in order to update the cluster's topology. */
+            redisClusterNode *prev_master = SLOT(c,c->redir_slot);
+            redisClusterNode *slave;
+            ZEND_HASH_FOREACH_PTR(prev_master->slaves, slave) {
+                if (slave == NULL) {
+                    continue;
+                }
+                if (!CLUSTER_REDIR_CMP(c, slave->sock)) {
+                    // Detected a failover, the redirected node was a replica
+                    // Remap the cluster's keyspace
+                    cluster_map_keyspace(c);
+                    return;
+                }
+            } ZEND_HASH_FOREACH_END();
+
             /* Create our node */
             node = cluster_node_create(c, c->redir_host, c->redir_host_len,
                 c->redir_port, c->redir_slot, 0);
@@ -1506,6 +1515,7 @@ PHP_REDIS_API int cluster_send_slot(redisCluster *c, short slot, char *cmd,
     if (cluster_sock_write(c, cmd, cmd_len, 1) == -1) {
         return -1;
     }
+    c->flags->txBytes += cmd_len;
 
     /* Check our response */
     if (cluster_check_response(c, &c->reply_type) != 0 ||
@@ -1543,7 +1553,7 @@ PHP_REDIS_API short cluster_send_command(redisCluster *c, short slot, const char
      * CLUSTERDOWN state from Redis Cluster. */
     do {
         /* Send MULTI to the socket if we're in MULTI mode but haven't yet */
-        if (c->flags->mode == MULTI && CMD_SOCK(c)->mode != MULTI) {
+        if (c->flags->mode == MULTI && c->cmd_sock->mode != MULTI) {
             /* We have to fail if we can't send MULTI to the node */
             if (cluster_send_multi(c, slot) == -1) {
                 CLUSTER_THROW_EXCEPTION("Unable to enter MULTI mode on requested slot", 0);
@@ -1589,11 +1599,13 @@ PHP_REDIS_API short cluster_send_command(redisCluster *c, short slot, const char
 
     // If we've detected the cluster is down, throw an exception
     if (c->clusterdown) {
+        cluster_cache_clear(c);
         CLUSTER_THROW_EXCEPTION("The Redis Cluster is down (CLUSTERDOWN)", 0);
         return -1;
     } else if (timedout || resp == -1) {
         // Make sure the socket is reconnected, it such that it is in a clean state
-        redis_sock_disconnect(c->cmd_sock, 1);
+        redis_sock_disconnect(c->cmd_sock, 1, 1);
+        cluster_cache_clear(c);
 
         if (timedout) {
             CLUSTER_THROW_EXCEPTION("Timed out attempting to find data in the correct node!", 0);
@@ -1665,27 +1677,33 @@ PHP_REDIS_API void cluster_bulk_resp(INTERNAL_FUNCTION_PARAMETERS, redisCluster
                               void *ctx)
 {
     char *resp;
+    zval z_unpacked, z_ret, *zv;
 
     // Make sure we can read the response
-    if (c->reply_type != TYPE_BULK ||
-       (resp = redis_sock_read_bulk_reply(c->cmd_sock, c->reply_len)) == NULL)
-    {
-        CLUSTER_RETURN_FALSE(c);
+    if (c->reply_type != TYPE_BULK) {
+        ZVAL_FALSE(&z_unpacked);
+        c->reply_len = 0;
+    } else if ((resp = redis_sock_read_bulk_reply(c->cmd_sock, c->reply_len)) == NULL) {
+        ZVAL_FALSE(&z_unpacked);
+    } else {
+        if (!redis_unpack(c->flags, resp, c->reply_len, &z_unpacked)) {
+            ZVAL_STRINGL_FAST(&z_unpacked, resp, c->reply_len);
+        }
+        efree(resp);
+    }
+
+    if (c->flags->flags & PHPREDIS_WITH_METADATA) {
+        redis_with_metadata(&z_ret, &z_unpacked, c->reply_len);
+        zv = &z_ret;
+    } else {
+        zv = &z_unpacked;
     }
 
     if (CLUSTER_IS_ATOMIC(c)) {
-        if (!redis_unpack(c->flags, resp, c->reply_len, return_value)) {
-            CLUSTER_RETURN_STRING(c, resp, c->reply_len);
-        }
+        RETVAL_ZVAL(zv, 0, 1);
     } else {
-        zval z_unpacked;
-        if (redis_unpack(c->flags, resp, c->reply_len, &z_unpacked)) {
-            add_next_index_zval(&c->multi_resp, &z_unpacked);
-        } else {
-            add_next_index_stringl(&c->multi_resp, resp, c->reply_len);
-        }
+        add_next_index_zval(&c->multi_resp, zv);
     }
-    efree(resp);
 }
 
 /* Bulk response where we expect a double */
@@ -1737,6 +1755,135 @@ PHP_REDIS_API void cluster_ping_resp(INTERNAL_FUNCTION_PARAMETERS, redisCluster
     CLUSTER_RETURN_BOOL(c, 1);
 }
 
+PHP_REDIS_API void
+cluster_hrandfield_resp(INTERNAL_FUNCTION_PARAMETERS, redisCluster *c, void *ctx)
+{
+    if (ctx == NULL) {
+        cluster_bulk_resp(INTERNAL_FUNCTION_PARAM_PASSTHRU, c, NULL);
+    } else if (ctx == PHPREDIS_CTX_PTR) {
+        cluster_mbulk_raw_resp(INTERNAL_FUNCTION_PARAM_PASSTHRU, c, NULL);
+    } else if (ctx == PHPREDIS_CTX_PTR + 1) {
+        return cluster_mbulk_zipdbl_resp(INTERNAL_FUNCTION_PARAM_PASSTHRU, c, NULL);
+    } else {
+        ZEND_ASSERT(!"memory corruption?");
+    }
+}
+
+PHP_REDIS_API void
+cluster_pop_resp(INTERNAL_FUNCTION_PARAMETERS, redisCluster *c, void *ctx)
+{
+    if (ctx == NULL) {
+        cluster_bulk_resp(INTERNAL_FUNCTION_PARAM_PASSTHRU, c, NULL);
+    } else if (ctx == PHPREDIS_CTX_PTR) {
+        cluster_mbulk_resp(INTERNAL_FUNCTION_PARAM_PASSTHRU, c, NULL);
+    } else {
+        ZEND_ASSERT(!"memory corruption?");
+    }
+}
+
+PHP_REDIS_API void
+cluster_lpos_resp(INTERNAL_FUNCTION_PARAMETERS, redisCluster *c, void *ctx)
+{
+    zval zret = {0};
+
+    c->cmd_sock->null_mbulk_as_null = c->flags->null_mbulk_as_null;
+    if (redis_read_lpos_response(&zret, c->cmd_sock, c->reply_type, c->reply_len, ctx) < 0) {
+        ZVAL_FALSE(&zret);
+    }
+
+    if (CLUSTER_IS_ATOMIC(c)) {
+        RETVAL_ZVAL(&zret, 0, 1);
+    } else {
+        add_next_index_zval(&c->multi_resp, &zret);
+    }
+}
+
+PHP_REDIS_API void
+cluster_geosearch_resp(INTERNAL_FUNCTION_PARAMETERS, redisCluster *c, void *ctx) {
+    zval zret = {0};
+
+    c->cmd_sock->null_mbulk_as_null = c->flags->null_mbulk_as_null;
+    if (c->reply_type != TYPE_MULTIBULK ||
+        redis_read_geosearch_response(&zret, c->cmd_sock, c->reply_len, ctx != NULL) < 0)
+    {
+        ZVAL_FALSE(&zret);
+    }
+
+    if (CLUSTER_IS_ATOMIC(c)) {
+        RETVAL_ZVAL(&zret, 0, 1);
+    } else {
+        add_next_index_zval(&c->multi_resp, &zret);
+    }
+}
+
+PHP_REDIS_API void
+cluster_zdiff_resp(INTERNAL_FUNCTION_PARAMETERS, redisCluster *c, void *ctx) {
+    if (ctx == NULL) {
+        cluster_mbulk_raw_resp(INTERNAL_FUNCTION_PARAM_PASSTHRU, c, NULL);
+    } else if (ctx == PHPREDIS_CTX_PTR) {
+        cluster_mbulk_zipdbl_resp(INTERNAL_FUNCTION_PARAM_PASSTHRU, c, NULL);
+    } else {
+        ZEND_ASSERT(!"memory corruption?");
+    }
+}
+
+PHP_REDIS_API void
+cluster_zadd_resp(INTERNAL_FUNCTION_PARAMETERS, redisCluster *c, void *ctx) {
+    ZEND_ASSERT(ctx == NULL || ctx == PHPREDIS_CTX_PTR);
+
+    if (ctx == NULL) {
+        cluster_long_resp(INTERNAL_FUNCTION_PARAM_PASSTHRU, c, NULL);
+    } else {
+        cluster_dbl_resp(INTERNAL_FUNCTION_PARAM_PASSTHRU, c, NULL);
+    }
+}
+
+
+PHP_REDIS_API void
+cluster_zrandmember_resp(INTERNAL_FUNCTION_PARAMETERS, redisCluster *c, void *ctx) {
+    if (ctx == NULL) {
+        cluster_bulk_resp(INTERNAL_FUNCTION_PARAM_PASSTHRU, c, NULL);
+    } else if (ctx == PHPREDIS_CTX_PTR) {
+        cluster_mbulk_raw_resp(INTERNAL_FUNCTION_PARAM_PASSTHRU, c, NULL);
+    } else if (ctx == PHPREDIS_CTX_PTR + 1) {
+        cluster_mbulk_zipdbl_resp(INTERNAL_FUNCTION_PARAM_PASSTHRU, c, NULL);
+    } else {
+        ZEND_ASSERT(!"memory corruption?");
+    }
+}
+
+PHP_REDIS_API void
+cluster_srandmember_resp(INTERNAL_FUNCTION_PARAMETERS, redisCluster *c, void *ctx) {
+    if (ctx == NULL) {
+        cluster_bulk_resp(INTERNAL_FUNCTION_PARAM_PASSTHRU, c, NULL);
+    } else if (ctx == PHPREDIS_CTX_PTR) {
+        cluster_mbulk_resp(INTERNAL_FUNCTION_PARAM_PASSTHRU, c, NULL);
+    } else {
+        ZEND_ASSERT(!"memory corruption?");
+    }
+}
+
+PHP_REDIS_API void
+cluster_object_resp(INTERNAL_FUNCTION_PARAMETERS, redisCluster *c, void *ctx) {
+    ZEND_ASSERT(ctx == PHPREDIS_CTX_PTR || ctx == PHPREDIS_CTX_PTR + 1);
+
+    if (ctx == PHPREDIS_CTX_PTR) {
+        cluster_long_resp(INTERNAL_FUNCTION_PARAM_PASSTHRU, c, NULL);
+    } else {
+        cluster_bulk_resp(INTERNAL_FUNCTION_PARAM_PASSTHRU, c, NULL);
+    }
+}
+
+PHP_REDIS_API void
+cluster_set_resp(INTERNAL_FUNCTION_PARAMETERS, redisCluster *c, void *ctx)
+{
+    if (ctx == NULL) {
+        cluster_bool_resp(INTERNAL_FUNCTION_PARAM_PASSTHRU, c, NULL);
+    } else {
+        cluster_bulk_resp(INTERNAL_FUNCTION_PARAM_PASSTHRU, c, NULL);
+    }
+}
+
 /* 1 or 0 response, for things like SETNX */
 PHP_REDIS_API void cluster_1_resp(INTERNAL_FUNCTION_PARAMETERS, redisCluster *c,
                            void *ctx)
@@ -1769,17 +1916,17 @@ PHP_REDIS_API void cluster_type_resp(INTERNAL_FUNCTION_PARAMETERS, redisCluster
     }
 
     // Switch on the type
-    if (strncmp (c->line_reply, "string", 6) == 0) {
+    if (redis_strncmp(c->line_reply, ZEND_STRL("string")) == 0) {
         CLUSTER_RETURN_LONG(c, REDIS_STRING);
-    } else if (strncmp(c->line_reply, "set", 3) == 0) {
+    } else if (redis_strncmp(c->line_reply, ZEND_STRL("set")) == 0) {
         CLUSTER_RETURN_LONG(c, REDIS_SET);
-    } else if (strncmp(c->line_reply, "list", 4) == 0) {
+    } else if (redis_strncmp(c->line_reply, ZEND_STRL("list")) == 0) {
         CLUSTER_RETURN_LONG(c, REDIS_LIST);
-    } else if (strncmp(c->line_reply, "hash", 4) == 0) {
+    } else if (redis_strncmp(c->line_reply, ZEND_STRL("hash")) == 0) {
         CLUSTER_RETURN_LONG(c, REDIS_HASH);
-    } else if (strncmp(c->line_reply, "zset", 4) == 0) {
+    } else if (redis_strncmp(c->line_reply, ZEND_STRL("zset")) == 0) {
         CLUSTER_RETURN_LONG(c, REDIS_ZSET);
-    } else if (strncmp(c->line_reply, "stream", 6) == 0) {
+    } else if (redis_strncmp(c->line_reply, ZEND_STRL("stream")) == 0) {
         CLUSTER_RETURN_LONG(c, REDIS_STREAM);
     } else {
         CLUSTER_RETURN_LONG(c, REDIS_NOT_FOUND);
@@ -1818,8 +1965,8 @@ PHP_REDIS_API void cluster_sub_resp(INTERNAL_FUNCTION_PARAMETERS, redisCluster *
 
     // Set up our callback pointers
     zval z_ret, z_args[4];
-    sctx->cb.retval = &z_ret;
-    sctx->cb.params = z_args;
+    sctx->cb.fci.retval = &z_ret;
+    sctx->cb.fci.params = z_args;
 
     /* We're in a subscribe loop */
     c->subscribed_slot = c->cmd_slot;
@@ -1838,9 +1985,9 @@ PHP_REDIS_API void cluster_sub_resp(INTERNAL_FUNCTION_PARAMETERS, redisCluster *
         }
 
         // Make sure we have a message or pmessage
-        if (!strncmp(Z_STRVAL_P(z_type), "message", 7) ||
-            !strncmp(Z_STRVAL_P(z_type), "pmessage", 8)
-        ) {
+        if (zend_string_equals_literal(Z_STR_P(z_type), "message") ||
+            zend_string_equals_literal(Z_STR_P(z_type), "pmessage"))
+        {
             is_pmsg = *Z_STRVAL_P(z_type) == 'p';
         } else {
             zval_dtor(&z_tab);
@@ -1872,12 +2019,10 @@ PHP_REDIS_API void cluster_sub_resp(INTERNAL_FUNCTION_PARAMETERS, redisCluster *
         }
 
         // Set arg count
-        sctx->cb.param_count = tab_idx;
+        sctx->cb.fci.param_count = tab_idx;
 
         // Execute our callback
-        if (zend_call_function(&(sctx->cb), &(sctx->cb_cache)) !=
-                              SUCCESS)
-        {
+        if (zend_call_function(&sctx->cb.fci, &sctx->cb.fci_cache) != SUCCESS) {
             break;
         }
 
@@ -2074,6 +2219,17 @@ cluster_variant_resp_generic(INTERNAL_FUNCTION_PARAMETERS, redisCluster *c,
     cluster_free_reply(r, 1);
 }
 
+PHP_REDIS_API void
+cluster_zrange_resp(INTERNAL_FUNCTION_PARAMETERS, redisCluster *c, void *ctx) {
+    cluster_cb cb;
+
+    ZEND_ASSERT(ctx == NULL || ctx == PHPREDIS_CTX_PTR);
+
+    cb = ctx ? cluster_mbulk_zipdbl_resp : cluster_mbulk_resp;
+
+    cb(INTERNAL_FUNCTION_PARAM_PASSTHRU, c, ctx);
+}
+
 PHP_REDIS_API void cluster_variant_resp(INTERNAL_FUNCTION_PARAMETERS, redisCluster *c,
                                         void *ctx)
 {
@@ -2131,8 +2287,9 @@ PHP_REDIS_API void cluster_gen_mbulk_resp(INTERNAL_FUNCTION_PARAMETERS,
 }
 
 /* HSCAN, SSCAN, ZSCAN */
-PHP_REDIS_API int cluster_scan_resp(INTERNAL_FUNCTION_PARAMETERS, redisCluster *c,
-                              REDIS_SCAN_TYPE type, long *it)
+PHP_REDIS_API int
+cluster_scan_resp(INTERNAL_FUNCTION_PARAMETERS, redisCluster *c,
+                  REDIS_SCAN_TYPE type, uint64_t *cursor)
 {
     char *pit;
 
@@ -2143,8 +2300,8 @@ PHP_REDIS_API int cluster_scan_resp(INTERNAL_FUNCTION_PARAMETERS, redisCluster *
     }
 
     // Read the BULK size
-    if (cluster_check_response(c, &c->reply_type),0 ||
-       c->reply_type != TYPE_BULK)
+    if (cluster_check_response(c, &c->reply_type) ||
+        c->reply_type != TYPE_BULK)
     {
         return FAILURE;
     }
@@ -2156,12 +2313,11 @@ PHP_REDIS_API int cluster_scan_resp(INTERNAL_FUNCTION_PARAMETERS, redisCluster *
     }
 
     // Push the new iterator value to our caller
-    *it = atol(pit);
+    *cursor = strtoull(pit, NULL, 10);
     efree(pit);
 
     // We'll need another MULTIBULK response for the payload
-    if (cluster_check_response(c, &c->reply_type) < 0)
-    {
+    if (cluster_check_response(c, &c->reply_type) < 0) {
         return FAILURE;
     }
 
@@ -2290,7 +2446,9 @@ cluster_xclaim_resp(INTERNAL_FUNCTION_PARAMETERS, redisCluster *c, void *ctx) {
 
     array_init(&z_msg);
 
-    if (redis_read_xclaim_response(c->cmd_sock, c->reply_len, &z_msg) < 0) {
+    ZEND_ASSERT(ctx == NULL || ctx == PHPREDIS_CTX_PTR);
+
+    if (redis_read_xclaim_reply(c->cmd_sock, c->reply_len, ctx == PHPREDIS_CTX_PTR, &z_msg) < 0) {
         zval_dtor(&z_msg);
         CLUSTER_RETURN_FALSE(c);
     }
@@ -2321,6 +2479,23 @@ cluster_xinfo_resp(INTERNAL_FUNCTION_PARAMETERS, redisCluster *c, void *ctx)
     add_next_index_zval(&c->multi_resp, &z_ret);
 }
 
+/* LMPOP, ZMPOP, BLMPOP, BZMPOP */
+PHP_REDIS_API void
+cluster_mpop_resp(INTERNAL_FUNCTION_PARAMETERS, redisCluster *c, void *ctx)
+{
+    zval z_ret;
+
+    c->cmd_sock->null_mbulk_as_null = c->flags->null_mbulk_as_null;
+    if (redis_read_mpop_response(c->cmd_sock, &z_ret, c->reply_len, ctx) == FAILURE) {
+        CLUSTER_RETURN_FALSE(c);
+    }
+
+    if (CLUSTER_IS_ATOMIC(c)) {
+        RETURN_ZVAL(&z_ret, 0, 0);
+    }
+    add_next_index_zval(&c->multi_resp, &z_ret);
+}
+
 static void
 cluster_acl_custom_resp(INTERNAL_FUNCTION_PARAMETERS, redisCluster *c, void *ctx,
                         int (*cb)(RedisSock*, zval*, long))
@@ -2384,8 +2559,9 @@ PHP_REDIS_API void cluster_multi_mbulk_resp(INTERNAL_FUNCTION_PARAMETERS,
                                      redisCluster *c, void *ctx)
 {
     zval *multi_resp = &c->multi_resp;
-    array_init(multi_resp);
+    uint8_t flags = c->flags->flags;
 
+    array_init(multi_resp);
     clusterFoldItem *fi = c->multi_head;
     while (fi) {
         /* Make sure our transaction didn't fail here */
@@ -2401,7 +2577,9 @@ PHP_REDIS_API void cluster_multi_mbulk_resp(INTERNAL_FUNCTION_PARAMETERS,
                 RETURN_FALSE;
             }
 
+            c->flags->flags = fi->flags;
             fi->callback(INTERNAL_FUNCTION_PARAM_PASSTHRU, c, fi->ctx);
+            c->flags->flags = flags;
         } else {
             /* Just add false */
             add_next_index_bool(multi_resp, 0);
@@ -2580,6 +2758,14 @@ cluster_mbulk_zipdbl_resp(INTERNAL_FUNCTION_PARAMETERS, redisCluster *c,
         mbulk_resp_loop_zipdbl, NULL);
 }
 
+PHP_REDIS_API void
+cluster_mbulk_dbl_resp(INTERNAL_FUNCTION_PARAMETERS, redisCluster *c,
+                       void *ctx)
+{
+    cluster_gen_mbulk_resp(INTERNAL_FUNCTION_PARAM_PASSTHRU, c,
+        mbulk_resp_loop_dbl, ctx);
+}
+
 /* Associate multi bulk response (for HMGET really) */
 PHP_REDIS_API void
 cluster_mbulk_assoc_resp(INTERNAL_FUNCTION_PARAMETERS, redisCluster *c,
@@ -2593,6 +2779,25 @@ cluster_mbulk_assoc_resp(INTERNAL_FUNCTION_PARAMETERS, redisCluster *c,
  * Various MULTI BULK reply callback functions
  */
 
+int mbulk_resp_loop_dbl(RedisSock *redis_sock, zval *z_result,
+                        long long count, void *ctx)
+{
+    char *line;
+    int line_len;
+
+    while (count--) {
+        line = redis_sock_read(redis_sock, &line_len);
+        if (line != NULL) {
+            add_next_index_double(z_result, atof(line));
+            efree(line);
+        } else {
+            add_next_index_bool(z_result, 0);
+        }
+    }
+
+    return SUCCESS;
+}
+
 /* MULTI BULK response where we don't touch the values (e.g. KEYS) */
 int mbulk_resp_loop_raw(RedisSock *redis_sock, zval *z_result,
                         long long count, void *ctx)
@@ -2801,7 +3006,7 @@ static zend_string **get_valid_seeds(HashTable *input, uint32_t *nseeds) {
         }
 
         /* Add as a key to avoid duplicates */
-        zend_hash_str_update_ptr(valid, Z_STRVAL_P(z_seed), Z_STRLEN_P(z_seed), NULL);
+        zend_hash_str_add_empty_element(valid, Z_STRVAL_P(z_seed), Z_STRLEN_P(z_seed));
     } ZEND_HASH_FOREACH_END();
 
     /* We need at least one valid seed */
@@ -2832,12 +3037,12 @@ cluster_validate_args(double timeout, double read_timeout, HashTable *seeds,
 {
     zend_string **retval;
 
-    if (timeout < 0L || timeout > INT_MAX) {
+    if (timeout > INT_MAX) {
         if (errstr) *errstr = "Invalid timeout";
         return NULL;
     }
 
-    if (read_timeout < 0L || read_timeout > INT_MAX) {
+    if (read_timeout > INT_MAX) {
         if (errstr) *errstr = "Invalid read timeout";
         return NULL;
     }
@@ -2903,13 +3108,11 @@ PHP_REDIS_API redisCachedCluster *cluster_cache_load(zend_string *hash) {
 
     if (le != NULL) {
         /* Sanity check on our list type */
-        if (le->type != le_cluster_slot_cache) {
-            php_error_docref(0, E_WARNING, "Invalid slot cache resource");
-            return NULL;
+        if (le->type == le_cluster_slot_cache) {
+            /* Success, return the cached entry */
+            return le->ptr;
         }
-
-        /* Success, return the cached entry */
-        return le->ptr;
+        php_error_docref(0, E_WARNING, "Invalid slot cache resource");
     }
 
     /* Not found */
@@ -2917,12 +3120,17 @@ PHP_REDIS_API redisCachedCluster *cluster_cache_load(zend_string *hash) {
 }
 
 /* Cache a cluster's slot information in persistent_list if it's enabled */
-PHP_REDIS_API int cluster_cache_store(zend_string *hash, HashTable *nodes) {
+PHP_REDIS_API void cluster_cache_store(zend_string *hash, HashTable *nodes) {
     redisCachedCluster *cc = cluster_cache_create(hash, nodes);
 
     redis_register_persistent_resource(cc->hash, cc, le_cluster_slot_cache);
+}
 
-    return SUCCESS;
+void cluster_cache_clear(redisCluster *c)
+{
+    if (c->cache_key) {
+        zend_hash_del(&EG(persistent_list), c->cache_key);
+    }
 }
 
 
